from transformers import FeatureExtractionMixin
from transformers import AutoFeatureExtractor
from transformers import BatchFeature
import pickle
from torchtext.data.utils import get_tokenizer
from torchtext.vocab import build_vocab_from_iterator, Vectors
from transformers.utils import cached_file
from pathlib import Path

#token_type: word, split, subword, char
#language: en, de, ko
class TorchtextFeatureExtractor(FeatureExtractionMixin):
    tokenizer = None
    vocab = None

    def __init__(
        self,
        token_type='word',
        language='en',
        min_freq=1,
        special_tokens=['<unk>', '<pad>'],
        default_token='<unk>',
        model_max_length=512,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.token_type = token_type
        self.language = language
        self.min_freq = min_freq
        self.special_tokens = special_tokens
        self.default_token = default_token
        self.vocab_size = None
        self.model_max_length = model_max_length

        if token_type == 'subword':
            TorchtextFeatureExtractor.tokenizer = get_tokenizer('subword')
        elif token_type == 'char':
            TorchtextFeatureExtractor.tokenizer = list 
        elif token_type == 'split':
            TorchtextFeatureExtractor.tokenizer = str.split 
        else: #word
            if language == 'ko':
                import konlpy
                okt = konlpy.tag.Okt() 
                TorchtextFeatureExtractor.tokenizer = okt.morphs
            else:  
                TorchtextFeatureExtractor.tokenizer = get_tokenizer('spacy', language=self.language)

    #padding: False, True, 'max_length'
    def __call__(self, texts, padding=False, vocab_path=None, return_tensors=None, **kwargs):
        if TorchtextFeatureExtractor.vocab == None and vocab_path != None:
            if Path(vocab_path).exists():
                vocab_file = f'{vocab_path}/vocab.pkl'
            else:
                #vocab_file = cached_file(path_or_repo_id='automatethem/imdb-text-classification', filename='vocab.pkl')
                vocab_file = cached_file(path_or_repo_id=vocab_path, filename='vocab.pkl')
                #print(vocab_file) #/root/.cache/huggingface/hub/models--automatethem--imdb-text-classification/snapshots/c588c4558da42a7c63fdb05bef68ecc35dc19710/vocab.pkl
            with open(vocab_file, 'rb') as f:
                TorchtextFeatureExtractor.vocab = pickle.load(f)
        
        if not isinstance(texts, list):
            texts = [texts]
        
        batch_max_length = 0
        if padding == False:
            pass
        else:
            if padding == 'max_length':
                pass
            else: 
                for text in texts:
                    tokens = self.tokenize(text)
                    ids = TorchtextFeatureExtractor.vocab(tokens)
                    if batch_max_length < len(ids):
                        batch_max_length = len(ids)

        input_ids = []
        for text in texts:
            print(text)
            tokens = self.tokenize(text)
            ids = TorchtextFeatureExtractor.vocab(tokens)
            if padding == False or padding == None:
                pass
            else:
                if padding == 'max_length':
                    max_length = self.model_max_length
                    ids = ids + ([self.get_token_to_id()['<pad>']] * (max_length - len(ids)))
                else:    
                    max_length = batch_max_length
                    ids = ids + ([self.get_token_to_id()['<pad>']] * (max_length - len(ids)))
            input_ids.append(ids)

        '''
        if return_tensors == 'pt':
            ids = torch.from_numpy(ids)
        return {'input_ids': input_ids}
        '''
        #''' 
        return BatchFeature(data={'input_ids': input_ids}, tensor_type=return_tensors)
        #'''
    
    def tokenize(self, text):
        tokens = TorchtextFeatureExtractor.tokenizer(text)
        return tokens

    def get_vocab(self):
        return TorchtextFeatureExtractor.vocab

    def get_token_to_id(self):
        return TorchtextFeatureExtractor.vocab.get_stoi()
    
    def get_id_to_token(self):
        return TorchtextFeatureExtractor.vocab.get_itos()
     
    def convert_tokens_to_ids(self, tokens):
        return [TorchtextFeatureExtractor.vocab(tokens)]

    def convert_ids_to_tokens(self, ids):
        id_to_token = self.get_id_to_token()
        return [id_to_token[id] for id in ids]

    def train_from_iterator(self, text_iterator):
        def tokens_iterator():
            for text in text_iterator:
                tokens = TorchtextFeatureExtractor.tokenizer(text)
                yield tokens
        vocab = build_vocab_from_iterator(tokens_iterator(), min_freq=self.min_freq, specials=self.special_tokens)
        #print(vocab)
        vocab.set_default_index(vocab[self.default_token]) # This index will be returned when OOV token is queried.
        self.vocab_size = len(vocab)
        #print(self.vocab_size) #204
        TorchtextFeatureExtractor.vocab = vocab

    #https://github.com/huggingface/transformers/blob/c8f35a9ce37bd03f37fcf8336172bdcbe7ffc86a/src/transformers/feature_extraction_utils.py#L333
    def save_pretrained(self, save_directory, push_to_hub=False, **kwargs):
        #print(save_directory) #/content/drive/MyDrive/models/pytorch/models/bank-loan-model-for-tabular-classification
        vocab_file = f'{save_directory}/vocab.pkl'
        with open(vocab_file, 'wb') as f:
            pickle.dump(TorchtextFeatureExtractor.vocab, f)
        return super().save_pretrained(save_directory, push_to_hub, **kwargs)
    
def register_auto():
    AutoFeatureExtractor.register(TorchtextFeatureExtractor, TorchtextFeatureExtractor)



from transformers import FeatureExtractionMixin
from transformers import AutoFeatureExtractor
from transformers import BatchFeature
import numpy as np
from sklearn.compose import make_column_transformer
from sklearn.preprocessing import OneHotEncoder
from sklearn.preprocessing import MinMaxScaler
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import LabelEncoder
from sklearn.preprocessing import LabelBinarizer
import pickle
from transformers.utils import cached_file
from pathlib import Path

class TabularFeatureExtractor(FeatureExtractionMixin):
    transformer = None
    le = None

    #labels_shape: list, one, onehot
    #labels_type: float, long
    def __init__(
        self,  
        continuous_columns=['ApplicantIncome', 'CoapplicantIncome', 'LoanAmount', 'Loan_Amount_Term', 'Credit_History'],
        categorical_columns=['Gender', 'Married', 'Dependents', 'Education', 'Self_Employed', 'Property_Area'],
        labels_column='Target',
        labels_shape='list',
        labels_type='float',
        **kwargs
    ):
        super().__init__(**kwargs)
        self.continuous_columns = continuous_columns
        self.categorical_columns = categorical_columns
        self.labels_column = labels_column
        self.labels_shape = labels_shape
        self.labels_type = labels_type
        self.labels = []

    def __call__(self, df, transformer_path=None, return_tensors=None, **kwargs):
        if TabularFeatureExtractor.transformer == None and transformer_path != None:
            if Path(transformer_path).exists():
                transformer_file = f'{transformer_path}/transformer.pkl'
            else:
                #transformer_file = cached_file(path_or_repo_id='automatethem/imdb-text-classification', filename='transformer.pkl')
                transformer_file = cached_file(path_or_repo_id=transformer_path, filename='transformer.pkl')
                #print(transformer_file) #/root/.cache/huggingface/hub/models--automatethem--iris-tabular-classification/snapshots/c588c4558da42a7c63fdb05bef68ecc35dc19710/transformer.pkl
            with open(transformer_file, 'rb') as f:
                TabularFeatureExtractor.transformer = pickle.load(f)
        
        x = df[self.continuous_columns + self.categorical_columns]
        x = TabularFeatureExtractor.transformer.transform(x)
        x = np.array(x, dtype=np.float32)
        #print(x.shape) #

        if self.labels_column not in df.columns:
            '''
            if return_tensors == 'pt':
                x = torch.from_numpy(x)
            return {'x': x}
            '''
            #'''
            return BatchFeature(data={'x': x}, tensor_type=return_tensors)
            #'''
        else:
            labels = df[self.labels_column]
            labels = labels.to_numpy()

            if df[self.labels_column].dtype == 'object':
                if self.labels_shape == 'list':
                    labels = TabularFeatureExtractor.le.transform(labels)
                elif self.labels_shape == 'one':
                    labels = TabularFeatureExtractor.le.transform(labels)
                    labels = labels.reshape(-1, 1)
                elif self.labels_shape == 'onehot':
                    labels = TabularFeatureExtractor.le.transform(labels)

            if self.labels_type == 'long':
                labels = np.array(labels, dtype=np.longlong)
            elif self.labels_type == 'float':
                labels = np.array(labels, dtype=np.float32)

            '''
            if return_tensors == 'pt':
                x = torch.from_numpy(x)
                labels = torch.from_numpy(labels)
            return {'x': x, 'labels': labels}
            '''
            #'''
            return BatchFeature(data={'x': x, 'labels': labels}, tensor_type=return_tensors)
            #'''

    def train(self, df):
        x = df[self.continuous_columns + self.categorical_columns]
        TabularFeatureExtractor.transformer = make_column_transformer(
            (OneHotEncoder(), self.categorical_columns),
            remainder='passthrough')
        if len(self.continuous_columns) >= 2:
            TabularFeatureExtractor.transformer = make_pipeline(TabularFeatureExtractor.transformer, MinMaxScaler())
        TabularFeatureExtractor.transformer.fit(x)

        if df[self.labels_column].dtype == 'object':
            labels = df[self.labels_column]
            if self.labels_shape == 'list':
                TabularFeatureExtractor.le = LabelEncoder()
                TabularFeatureExtractor.le.fit(labels)
                #print(TabularFeatureExtractor.le.classes_) #['N' 'Y']
                self.labels = TabularFeatureExtractor.le.classes_ 
            elif self.labels_shape == 'one':
                TabularFeatureExtractor.le = LabelEncoder()
                TabularFeatureExtractor.le.fit(labels)
                #print(TabularFeatureExtractor.le.classes_) #['N' 'Y']
                self.labels = TabularFeatureExtractor.le.classes_ 
            elif self.labels_shape == 'onehot':
                TabularFeatureExtractor.le = LabelBinarizer()
                labels = df[self.labels_column]
                TabularFeatureExtractor.le.fit(labels)
                #print(TabularFeatureExtractor.le.classes_) #['N' 'Y']
                self.labels = TabularFeatureExtractor.le.classes_ 

    #https://github.com/huggingface/transformers/blob/c8f35a9ce37bd03f37fcf8336172bdcbe7ffc86a/src/transformers/feature_extraction_utils.py#L333
    def save_pretrained(self, save_directory, push_to_hub=False, **kwargs):
        #print(save_directory) #/content/drive/MyDrive/models/pytorch/models/bank-loan-model-for-tabular-classification
        transformer_file = f'{save_directory}/transformer.pkl'
        with open(transformer_file, 'wb') as f:
            pickle.dump(TabularFeatureExtractor.transformer, f)
        return super().save_pretrained(save_directory, push_to_hub, **kwargs)

def register_auto():
    AutoFeatureExtractor.register(TabularFeatureExtractor, TabularFeatureExtractor)



from transformers import PretrainedConfig
from transformers import PreTrainedModel
from transformers import AutoModel
from transformers import AutoTokenizer
from transformers import AutoConfig
from transformers import AutoModelForSequenceClassification
import torch
from torch.nn import functional as F
import pytorch_supporter
import transformers
from .configuration_custom_bert import CustomBertConfig

'''
class CustomBertForSequenceClassification(PreTrainedModel):
    config_class = CustomBertConfig

    def __init__(self, config):
        super().__init__(config)
        self.bert_layer = AutoModel.from_pretrained(config.bert_model_path)
        self.linear_layer = torch.nn.Linear(in_features=768, out_features=config.num_labels)

    def forward(self, input_ids, attention_mask=None, token_type_ids=None, labels=None):
        outputs = self.bert_layer(input_ids=input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids)
        #print(outputs.keys()) #dict_keys(['last_hidden_state', 'pooler_output'])
        pooler_output = outputs.pooler_output
        #print(pooler_output.shape) #torch.Size([1, 768])
        logits = self.linear_layer(pooler_output)
        #print(logits.shape) #torch.Size([16, 3])
        if labels == None:
            return transformers.file_utils.ModelOutput({'logits': logits})
        else:
            loss = F.nll_loss(F.log_softmax(logits), labels) #원핫 벡터를 넣을 필요없이 바로 실제값을 인자로 사용 #nll은 Negative Log Likelihood의 약자
            return transformers.file_utils.ModelOutput({'loss': loss, 'logits': logits})
'''
#'''
#torch.nn.Sequential 버전
class CustomBertForSequenceClassification(PreTrainedModel):
    config_class = CustomBertConfig

    def __init__(self, config):
        super().__init__(config)
        self.layer = torch.nn.Sequential(
            pytorch_supporter.layers.DictToParameters(AutoModel.from_pretrained(config.bert_model_path)),
            pytorch_supporter.layers.SelectFromArray(index=1), #x.pooler_output
            torch.nn.Linear(in_features=768, out_features=config.num_labels)
        )

    def forward(self, input_ids, attention_mask=None, token_type_ids=None, labels=None):
        logits = self.layer(dict(input_ids=input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids))
        #print(pooler_output.shape) #torch.Size([1, 768])
        #print(logits.shape) #torch.Size([16, 3])
        if labels == None:
            return transformers.file_utils.ModelOutput({'logits': logits})
        else:
            loss = F.nll_loss(F.log_softmax(logits), labels) #원핫 벡터를 넣을 필요없이 바로 실제값을 인자로 사용 #nll은 Negative Log Likelihood의 약자
            return transformers.file_utils.ModelOutput({'loss': loss, 'logits': logits})
#'''

def register_auto():
    AutoModelForSequenceClassification.register(CustomBertConfig, CustomBertForSequenceClassification)



from transformers import PretrainedConfig
from transformers import AutoConfig

class FasterRcnnConfig(PretrainedConfig):
    model_type = "faster-rcnn"

    def __init__(self, **kwargs):
        #https://github.com/huggingface/transformers/blob/98d88b23f54e5a23e741833f1e973fdf600cc2c5/src/transformers/configuration_utils.py#L323
        #Keys are always strings in JSON so convert ids to int here.       
        super().__init__(**kwargs)   

def register_auto():
    AutoConfig.register("faster-rcnn", FasterRcnnConfig)



from transformers import Pipeline
from transformers.pipelines import PIPELINE_REGISTRY

class TabularRegressionPipeline(Pipeline):
    def _sanitize_parameters(self, **kwargs):
        preprocess_kwargs = {}
        if "transformer_path" in kwargs:
            preprocess_kwargs["transformer_path"] = kwargs["transformer_path"]
        postprocess_kwargs = {}
        return preprocess_kwargs, {}, postprocess_kwargs

    def preprocess(self, inputs, transformer_path=None):
        return self.feature_extractor(inputs, transformer_path=transformer_path, return_tensors=self.framework)        

    def _forward(self, model_inputs):
        return self.model(**model_inputs)

    def postprocess(self, model_outputs):
        logits = model_outputs['logits']
        #print(logits.shape) #torch.Size([1, 1])
        postprocessed = []
        for logit in logits:
            postprocessed.append({"logit": logit[0].item()})        
        return postprocessed

def register_pipeline():
    PIPELINE_REGISTRY.register_pipeline('tabular-regression', 
                                    #pt_model=AutoModelForTabularRegression
                                    pipeline_class=TabularRegressionPipeline)



from transformers import PretrainedConfig
from transformers import PreTrainedModel
from transformers import AutoModel
from transformers import AutoTokenizer
from transformers import AutoConfig
from transformers import AutoModelForSequenceClassification
import torch
from torch.nn import functional as F
import pytorch_supporter
import transformers
from .configuration_embedded_rnn import EmbeddedRnnConfig

class EmbeddedRnnForSequenceClassification(PreTrainedModel):
    config_class = EmbeddedRnnConfig

    def __init__(self, config):
        super().__init__(config)        
        self.layer = torch.nn.Sequential(
            torch.nn.Embedding(num_embeddings=config.vocab_size, embedding_dim=32),
            pytorch_supporter.layers.HiddenStateResetLSTM(input_size=32, hidden_size=32, batch_first=True),
            pytorch_supporter.layers.LSTMLastHiddenState(),
            torch.nn.Linear(in_features=32, out_features=config.num_labels)
        )

    def forward(self, input_ids, labels=None):
        logits = self.layer(input_ids)
        #print(logits.shape) #torch.Size([16, 3])
        if labels == None:
            return transformers.file_utils.ModelOutput({'logits': logits})
        else:
            loss = F.nll_loss(F.log_softmax(logits), labels) #원핫 벡터를 넣을 필요없이 바로 실제값을 인자로 사용 #nll은 Negative Log Likelihood의 약자
            return transformers.file_utils.ModelOutput({'loss': loss, 'logits': logits})

from transformers import PretrainedConfig
from transformers import PreTrainedModel
from transformers import AutoModel
from transformers import AutoTokenizer
from transformers import AutoConfig
import torch
from torch.nn import functional as F
import pytorch_supporter
import transformers
from .configuration_embedded_rnn import EmbeddedRnnConfig

class EmbeddedRnnForFixedLengthTranslation(PreTrainedModel):
    config_class = EmbeddedRnnConfig

    def __init__(self, config):
        super().__init__(config)        
        self.layer = torch.nn.Sequential(
            torch.nn.Embedding(num_embeddings=config.vocab_size, embedding_dim=32),
            pytorch_supporter.layers.HiddenStateResetLSTM(input_size=32, hidden_size=32, batch_first=True),
            pytorch_supporter.layers.SelectFromArray(index=0), #lstm output, 모든 타임 스텝(토큰: 숫자, 문자, 단어)의 숨은 상태, torch.Size([8, 380, 32]) 
            torch.nn.Linear(in_features=32, out_features=config.vocab_size)
        )

    def forward(self, input_ids, labels=None):
        logits = self.layer(input_ids)
        #print(logits.shape) #torch.Size([2, 3, 7])
        if labels == None:
            return transformers.file_utils.ModelOutput({'logits': logits})
        else:
            #print(labels.shape) #torch.Size([2, 3])
            logits_ = logits.view(-1, logits.size(-1))
            labels_ = labels.view(-1)
            #print(logits_.shape) #torch.Size([6, 7])
            #print(labels_.shape) #torch.Size([6])
            loss = F.nll_loss(F.log_softmax(logits_), labels_) #원핫 벡터를 넣을 필요없이 바로 실제값을 인자로 사용 #nll은 Negative Log Likelihood의 약자
            return transformers.file_utils.ModelOutput({'loss': loss, 'logits': logits})

def register_auto():
    AutoModelForSequenceClassification.register(EmbeddedRnnConfig, EmbeddedRnnForSequenceClassification)
    #AutoModelForFixedLengthTranslation.register(EmbeddedRnnConfig, EmbeddedRnnForFixedLengthTranslation)
    



from transformers import PretrainedConfig
from transformers import PreTrainedModel
from transformers import AutoModel
from transformers import AutoTokenizer
from transformers import AutoConfig
from transformers import AutoModelForImageClassification
import torch
from torch.nn import functional as F
import transformers

class CnnConfig(PretrainedConfig):
    model_type = "cnn"

    def __init__(self, **kwargs):
        #https://github.com/huggingface/transformers/blob/98d88b23f54e5a23e741833f1e973fdf600cc2c5/src/transformers/configuration_utils.py#L323
        #Keys are always strings in JSON so convert ids to int here.       
        super().__init__(**kwargs) 

def register_auto():
    AutoConfig.register("cnn", CnnConfig)



from . import models
from . import pipelines

#

'''
from .models.ann import configuration_ann
from .models.ann import feature_extraction_ann
from .models.ann import modeling_ann

from .models.cnn import configuration_cnn
from .models.cnn import modeling_cnn
from .models.cnn import image_processing_cnn

from .models.custom_bert import configuration_custom_bert
from .models.custom_bert import modeling_custom_bert

from .models.embedded_1dcnn import configuration_embedded_1dcnn
from .models.embedded_1dcnn import modeling_embedded_1dcnn

from .models.embedded_rnn import configuration_embedded_rnn
from .models.embedded_rnn import modeling_embedded_rnn
from .models.embedded_rnn import feature_extraction_embedded_rnn

from .models.pretrained_embedded_rnn import configuration_pretrained_embedded_rnn
from .models.pretrained_embedded_rnn import modeling_pretrained_embedded_rnn

from .models.faster_rcnn import configuration_faster_rcnn
from .models.faster_rcnn import modeling_faster_rcnn
from .models.faster_rcnn import image_processing_faster_rcnn

from .models.rnn import configuration_rnn
from .models.rnn import modeling_rnn
from .models.rnn import feature_extraction_rnn

def register_auto():
    configuration_ann.register_auto()
    modeling_ann.register_auto()
    feature_extraction_ann.register_auto()

    configuration_cnn.register_auto()
    modeling_cnn.register_auto()
    image_processing_cnn.register_auto()

    configuration_custom_bert.register_auto()
    modeling_custom_bert.register_auto()

    configuration_embedded_1dcnn.register_auto()
    modeling_embedded_1dcnn.register_auto()

    configuration_embedded_rnn.register_auto()
    modeling_embedded_rnn.register_auto()
    feature_extraction_embedded_rnn.register_auto()

    configuration_pretrained_embedded_rnn.register_auto()
    modeling_pretrained_embedded_rnn.register_auto()

    configuration_faster_rcnn.register_auto()
    modeling_faster_rcnn.register_auto()
    image_processing_faster_rcnn.register_auto()

    configuration_rnn.register_auto()
    modeling_rnn.register_auto()
    feature_extraction_rnn.register_auto()
    
from .pipelines import tabular_regression
from .pipelines import tabular_classification
from .pipelines import image_classification
from .pipelines import fixed_length_translation

def register_pipeline():
    tabular_regression.register_pipeline()
    tabular_classification.register_pipeline()
    image_classification.register_pipeline()
    fixed_length_translation.register_pipeline()
'''



from . import configuration_embedded_rnn
from . import modeling_embedded_rnn
from . import feature_extraction_embedded_rnn

def register_auto():
    configuration_embedded_rnn.register_auto()
    modeling_embedded_rnn.register_auto()
    feature_extraction_embedded_rnn.register_auto()



from transformers import PretrainedConfig
from transformers import PreTrainedModel
from transformers import AutoModel
from transformers import AutoTokenizer
from transformers import AutoConfig
from transformers import AutoModelForSequenceClassification
from .configuration_pretrained_embedded_rnn import PretrainedEmbeddedRnnConfig
import torch
from torch.nn import functional as F
from torchtext.vocab import build_vocab_from_iterator, Vectors
import transformers
import pytorch_supporter

class PretrainedEmbeddedRnnForSequenceClassification(PreTrainedModel):
    config_class = PretrainedEmbeddedRnnConfig

    def __init__(self, config):
        super().__init__(config)  
        self.layer = torch.nn.Sequential(
            self.load_pretrained_embedding(num_embeddings=config.vocab_size, embedding_dim=300, id_to_token=config.id_to_token),
            pytorch_supporter.layers.HiddenStateResetLSTM(input_size=300, hidden_size=32, batch_first=True),
            pytorch_supporter.layers.LSTMLastHiddenState(),
            torch.nn.Linear(in_features=32, out_features=config.num_labels)
        )

    def load_pretrained_embedding(self, num_embeddings, embedding_dim, id_to_token):
        #사이즈가 작은 파일들만 가져왔습니다. 다른 모델을 써보고 싶다면, 아래 링크를 참고해서 코드를 변경해서 사용하세요.
        #https://pytorch.org/text/stable/_modules/torchtext/vocab/vectors.html#Vectors
        '''
        #glove
        pretrained_vectors = Vectors(name='glove.6B.300d.txt', url='http://nlp.stanford.edu/data/glove.6B.zip')
        pretrained_emb = pretrained_vectors.get_vecs_by_tokens(id_to_token, lower_case_backup=True)
        embedding = torch.nn.Embedding(num_embeddings=num_embeddings, embedding_dim=embedding_dim).from_pretrained(pretrained_emb, freeze=False)
        return embedding
        '''
        #'''
        #fasttext
        pretrained_vectors = Vectors(name='wiki.simple.vec', url='https://dl.fbaipublicfiles.com/fasttext/vectors-wiki/wiki.simple.vec')
        pretrained_emb = pretrained_vectors.get_vecs_by_tokens(id_to_token, lower_case_backup=True)
        embedding = torch.nn.Embedding(num_embeddings=num_embeddings, embedding_dim=embedding_dim).from_pretrained(pretrained_emb, freeze = False)
        return embedding
        #'''

    def forward(self, input_ids, labels=None):
        logits = self.layer(input_ids)
        #print(logits.shape) #torch.Size([16, 3])
        if labels == None:
            return transformers.file_utils.ModelOutput({'logits': logits})
        else:
            loss = F.nll_loss(F.log_softmax(logits), labels) #원핫 벡터를 넣을 필요없이 바로 실제값을 인자로 사용 #nll은 Negative Log Likelihood의 약자
            return transformers.file_utils.ModelOutput({'loss': loss, 'logits': logits})

def register_auto():
    AutoModelForSequenceClassification.register(PretrainedEmbeddedRnnConfig, PretrainedEmbeddedRnnForSequenceClassification)



#https://huggingface.co/docs/transformers/main_classes/pipelines#transformers.ImageClassificationPipeline
#https://github.com/huggingface/transformers/blob/v4.30.0/src/transformers/pipelines/image_classification.py#L32
from transformers import Pipeline
from transformers.pipelines import PIPELINE_REGISTRY
from torch.nn import functional as F

class ImageClassificationPipeline(Pipeline):
    def _sanitize_parameters(self, **kwargs):
        preprocess_kwargs = {}
        postprocess_kwargs = {}
        if "top_k" in kwargs:
            preprocess_kwargs["top_k"] = kwargs["top_k"]
        return preprocess_kwargs, {}, postprocess_kwargs

    def preprocess(self, inputs):
        return self.image_processor(inputs, return_tensors=self.framework)        

    def _forward(self, model_inputs):
        return self.model(**model_inputs)

    def postprocess(self, model_outputs, top_k=5):
        logits = model_outputs['logits']
        probabilities = F.softmax(logits, dim=-1)
        results = []
        for i, probability in enumerate(probabilities[0]):
            label = self.model.config.id2label[i]
            results.append({'label': label, 'score': probability.item()})   
        results.sort(key=lambda x: x['score'], reverse=True)
        if top_k != None:
            results = results[:top_k]            
        return results

def register_pipeline():
    PIPELINE_REGISTRY.register_pipeline('image-classification', 
                                    #pt_model=AutoModelForImageClassification,
                                    pipeline_class=ImageClassificationPipeline)



from . import configuration_pretrained_embedded_rnn
from . import modeling_pretrained_embedded_rnn

def register_auto():
    configuration_pretrained_embedded_rnn.register_auto()
    modeling_pretrained_embedded_rnn.register_auto()



from . import configuration_rnn
from . import modeling_rnn
from . import feature_extraction_rnn

def register_auto():
    configuration_rnn.register_auto()
    modeling_rnn.register_auto()
    feature_extraction_rnn.register_auto()



from . import image_classification
from . import tabular_classification
from . import tabular_regression
from . import fixed_length_translation

from .image_classification import ImageClassificationPipeline
from .tabular_classification import TabularClassificationPipeline
from .tabular_regression import TabularRegressionPipeline
from .fixed_length_translation import FixedLengthTranslationPipeline



from . import configuration_ann
from . import modeling_ann
from . import feature_extraction_ann

def register_auto():
    configuration_ann.register_auto()
    modeling_ann.register_auto()
    feature_extraction_ann.register_auto()
  



from transformers import FeatureExtractionMixin
import torch
from transformers.image_utils import is_batched
from transformers import BatchFeature
from transformers import AutoFeatureExtractor
from torchvision import transforms

#Custom Feature extractor class for Wav2Vec2 (https://github.com/huggingface/transformers/blob/main/src/transformers/models/wav2vec2/feature_extraction_wav2vec2.py)
class Wav2Vec2FeatureExtractor(FeatureExtractionMixin):
    def __init__(self, sampling_rate=16000, **kwargs):
        super().__init__(**kwargs)
        self.sampling_rate = sampling_rate

    def __call__(self, raw_speeches, sampling_rate=None, return_tensors=None, **kwargs):
        if sampling_rate is not None:
            if sampling_rate != self.sampling_rate:
                raise ValueError(
                    f"The model corresponding to this feature extractor: {self} was trained using a sampling rate of"
                    f" {self.sampling_rate}. Please make sure that the provided `raw_speech` input was sampled with"
                    f" {self.sampling_rate} and not {sampling_rate}."
                )
        else:
            pint(
                "It is strongly recommended to pass the ``sampling_rate`` argument to this function. "
                "Failing to do so can result in silent errors that might be hard to debug."
            )
        is_batched = bool(
            isinstance(raw_speeches, (list, tuple))
            and (isinstance(raw_speeches[0], np.ndarray) or isinstance(raw_speeches[0], (tuple, list)))
        )
        if not is_batched:
            raw_speeches = [raw_speeches]
        raw_speeches = self.normalize(raw_speeches)
        #print(raw_speeches[0].shape) #
        '''
        if return_tensors == 'pt':
            raw_speeches = torch.stack(raw_speeches)
        return {"input_values": raw_speeches}
        '''
        #'''
        return BatchFeature(data={"input_values": raw_speeches}, tensor_type=return_tensors)
        #'''

    def normalize(self, input_values):
        #print(input_values) #[array([-0.00234106, -0.00490547, -0.00566588, ..., -0.00049007, -0.00182217, -0.00376346], dtype=float32)]
        #https://github.com/huggingface/transformers/blob/94b3f544a1f5e04b78d87a2ae32a7ac252e22e31/src/transformers/models/wav2vec2/feature_extraction_wav2vec2.py#L98
        normed_input_values = [(x - x.mean()) / np.sqrt(x.var() + 1e-7) for x in input_values]
        #print(normed_input_values) #[array([-0.02099453, -0.0440244 , -0.0508533 , ..., -0.00437161, -0.01633461, -0.03376846], dtype=float32)]
        return normed_input_values

def register_auto():
    AutoFeatureExtractor.register(Wav2Vec2FeatureExtractor, Wav2Vec2FeatureExtractor)



from . import configuration_cnn
from . import modeling_cnn
from . import image_processing_cnn

def register_auto():
    configuration_cnn.register_auto()
    modeling_cnn.register_auto()
    image_processing_cnn.register_auto()



from . import configuration_faster_rcnn
from . import modeling_faster_rcnn
from . import image_processing_faster_rcnn

def register_auto():
    configuration_faster_rcnn.register_auto()
    modeling_faster_rcnn.register_auto()
    image_processing_faster_rcnn.register_auto()



from transformers import ImageProcessingMixin
from PIL import Image
import torch
from transformers.image_utils import is_batched
from transformers import BatchFeature
from transformers import AutoImageProcessor
from torchvision import transforms
from torch.nn import functional as F

class FasterRcnnImageProcessor(ImageProcessingMixin):
    def __init__(self, image_mean=[0.5, 0.5, 0.5], image_std=[0.5, 0.5, 0.5], **kwargs):
        super().__init__(**kwargs)
        self.image_mean = image_mean
        self.image_std = image_std

    def __call__(self, images, annotations=None, return_tensors=None, **kwargs):
        return_tensors = None #return_tensors 미지원
        composed_transforms = transforms.Compose([
            transforms.ToTensor() #최소 최대 정규화 ([0, 1])
        ])
        if not is_batched(images):
            images = [images]
        images = [image.convert('RGB') for image in images]
        images = [composed_transforms(image) for image in images]
        #print(images[0].shape) #torch.Size([3, 28, 28])
        if annotations == None:
            '''
            if return_tensors == 'pt':
                images = torch.stack(images)
            return {'pixel_values': images}
            '''
            #'''
            return BatchFeature(data={'pixel_values': images}, tensor_type=return_tensors)
            #'''
        else:
            if not isinstance(annotations, list):
                annotations = [annotations]
            labels = []
            for sub_annotations in annotations:
                label = self.get_label(sub_annotations)
                labels.append(label)
            '''
            if return_tensors == 'pt':
                images = torch.stack(images)
            return {'pixel_values': images, 'labels': labels}
            '''
            #'''
            return BatchFeature(data={'pixel_values': images, 'labels': labels}, tensor_type=return_tensors)
            #'''

    def get_label(self, sub_annotations):
        #print(sub_annotations) #[{'bbox': [34.5299987793, 556.8300170898, 401.4400024414, 276.2600097656], 'category_id': 0}]
        boxes = []
        labels = []
        for sub_annotation in sub_annotations:
            #print(annotation) #{"bbox":[34.5299987793, 556.8300170898, 401.4400024414, 276.2600097656], "category_id":0}
            x = sub_annotation['bbox'][0] #데이타셋 (coco) #x, y, w, h
            y = sub_annotation['bbox'][1]
            w = sub_annotation['bbox'][2]
            h = sub_annotation['bbox'][3]
            sub_annotation['bbox'][0] = x #x1 #모델 (pascal_voc) #min_x, min_y, max_x, max_y
            sub_annotation['bbox'][1] = y #x2
            sub_annotation['bbox'][2] = x + w #x2
            sub_annotation['bbox'][3] = y + h #y2
            boxes.append(sub_annotation['bbox'])
            labels.append(sub_annotation['category_id'] + 1) #torchvision.models.detection.fasterrcnn_resnet50_fpn 에서는 0 (백그라운드), 1 부터        
        return {'boxes': boxes, 'labels': labels} 

    def post_process_object_detection(self, outputs, threshold=0.5, target_sizes=None):
        results = []
        for boxes, labels, scores in zip(outputs['boxes'], outputs['labels'], outputs['scores']):
            boxes_ = boxes[scores > threshold]
            labels_ = labels[scores > threshold]
            scores_ = scores[scores > threshold]
            results.append({"scores": scores_, "labels": labels_, "boxes": boxes_})
        return results

def register_auto():
    AutoImageProcessor.register(FasterRcnnImageProcessor, FasterRcnnImageProcessor)
